@@ -20,10 +20,13 @@ Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 
 #include <Python.h>
 #include <structmember.h>
+#include <datetime.h>
 
 typedef struct {
     PyObject_HEAD
     sd_journal *j;
+    PyObject *default_call;
+    PyObject *call_dict;
 } Journalctl;
 static PyTypeObject JournalctlType;
 
@@ -31,25 +34,108 @@ static void
 Journalctl_dealloc(Journalctl* self)
 {
     sd_journal_close(self->j);
+    Py_XDECREF(self->default_call);
+    Py_XDECREF(self->call_dict);
     Py_TYPE(self)->tp_free((PyObject*)self);
 }
 
+static PyObject *
+Journalctl_new(PyTypeObject *type, PyObject *args, PyObject *kwds)
+{
+    Journalctl *self;
+
+    self = (Journalctl *)type->tp_alloc(type, 0);
+    if (self != NULL) {
+        PyObject *globals, *temp;
+
+        globals = PyEval_GetBuiltins();
+        temp = PyImport_ImportModule("functools");
+        PyDict_SetItemString(globals, "functools", temp);
+        Py_DECREF(temp);
+        temp = PyImport_ImportModule("datetime");
+        PyDict_SetItemString(globals, "datetime", temp);
+        Py_DECREF(temp);
+
+#if PY_MAJOR_VERSION >=3
+        self->default_call = PyRun_String("functools.partial(str, encoding='utf-8')", Py_eval_input, globals, NULL);
+#else
+        self->default_call = PyRun_String("functools.partial(unicode, encoding='utf-8')", Py_eval_input, globals, NULL);
+#endif
+
+        self->call_dict = PyRun_String("{"
+            "'PRIORITY': int,"
+            "'_UID': int,"
+            "'_GID': int,"
+            "'_PID': int,"
+            "'SYSLOG_FACILITY': int,"
+            "'SYSLOG_PID': int,"
+            "'_AUDIT_SESSION': int,"
+            "'_AUDIT_LOGINUID': int,"
+            "'_SYSTEMD_SESSION': int,"
+            "'_SYSTEMD_OWNER_UID': int,"
+            "'CODE_LINE': int,"
+            "'ERRNO': int,"
+            "'_SOURCE_REALTIME_TIMESTAMP': lambda x: datetime.datetime.fromtimestamp(float(x)/1E6),"
+            "'__REALTIME_TIMESTAMP': lambda x: datetime.datetime.fromtimestamp(float(x)/1E6),"
+            "'__MONOTONIC_TIMESTAMP': lambda x: datetime.timedelta(microseconds=float(x)),"
+            "}", Py_eval_input, globals, NULL);
+    }
+
+    return (PyObject *) self;
+}
+
 PyDoc_STRVAR(Journalctl__doc__,
-"Journalctl([flags]) -> Journalctl instance\n\n"
+"Journalctl([flags][, default_call][, call_dict]) -> Journalctl instance\n\n"
 "Returns instance of Journalctl, which allows filtering and return\n"
 "of journal entries.\n"
 "Argument `flags` sets open flags of the journal, which can be one\n"
 "of, or ORed combination of: SD_JOURNAL_LOCAL_ONLY (default) opens\n"
 "journal on local machine only; SD_JOURNAL_RUNTIME_ONLY opens only\n"
 "volatile journal files; and SD_JOURNAL_SYSTEM_ONLY opens only\n"
-"journal files of system services and the kernel.");
+"journal files of system services and the kernel.\n"
+"Argument `default_call` must be a callable that accepts one\n"
+"argument which is string/bytes value of a field and returns\n"
+"python object.\n"
+"Argument `call_dict` is a dictionary where the key represents\n"
+"a field name, and value is a callable as per `default_call`.\n"
+"A set of sane defaults for `default_call` and `call_dict` are\n"
+"present.");
 static int
-Journalctl_init(Journalctl *self, PyObject *args)
+Journalctl_init(Journalctl *self, PyObject *args, PyObject *keywds)
 {
     int flags=SD_JOURNAL_LOCAL_ONLY;
-    if (! PyArg_ParseTuple(args, "|i", &flags))
+    PyObject *default_call=NULL, *call_dict=NULL;
+
+    static char *kwlist[] = {"flags", "default_call", "call_dict", NULL};
+    if (! PyArg_ParseTupleAndKeywords(args, keywds, "|iOO", kwlist,
+                                      &flags, &default_call, &call_dict))
         return 1;
 
+    if (default_call) {
+        if (PyCallable_Check(default_call) || default_call == Py_None) {
+            Py_DECREF(self->default_call);
+            self->default_call = default_call;
+            Py_INCREF(self->default_call);
+        }else{
+            PyErr_SetString(PyExc_TypeError, "Default call not callable");
+            return 1;
+        }
+    }
+
+    if (call_dict) {
+        if (PyDict_Check(call_dict)) {
+            Py_DECREF(self->call_dict);
+            self->call_dict = call_dict;
+            Py_INCREF(self->call_dict);
+        }else if (call_dict == Py_None) {
+            Py_DECREF(self->call_dict);
+            self->call_dict = PyDict_New();
+        }else{
+            PyErr_SetString(PyExc_TypeError, "Call dictionary must be dict type");
+            return 1;
+        }
+    }
+
     int r;
     r = sd_journal_open(&self->j, flags);
     if (r < 0) {
@@ -60,6 +146,42 @@ Journalctl_init(Journalctl *self, PyObject *args)
     return 0;
 }
 
+static PyObject *
+Journalctl___process_field(Journalctl *self, PyObject *key, const void *value, ssize_t value_len)
+{
+    PyObject *callable=NULL, *return_value=NULL;
+    if (PyDict_Check(self->call_dict))
+        callable = PyDict_GetItem(self->call_dict, key);
+
+    if (callable) {
+#if PY_MAJOR_VERSION >=3
+        return_value = PyObject_CallFunction(callable, "y#", value, value_len);
+#else
+        return_value = PyObject_CallFunction(callable, "s#", value, value_len);
+#endif
+        if (!return_value)
+            PyErr_Clear();
+    }
+    if (!return_value)
+#if PY_MAJOR_VERSION >=3
+        return_value = PyObject_CallFunction(self->default_call, "y#", value, value_len);
+#else
+        return_value = PyObject_CallFunction(self->default_call, "s#", value, value_len);
+#endif
+    if (!return_value) {
+        PyErr_Clear();
+#if PY_MAJOR_VERSION >=3
+        return_value = PyBytes_FromStringAndSize(value, value_len);
+#else
+        return_value = PyString_FromStringAndSize(value, value_len);
+#endif
+    }
+    if (!return_value) {
+        return_value = Py_None;
+    }
+    return return_value;
+}
+
 PyDoc_STRVAR(Journalctl_get_next__doc__,
 "get_next([skip]) -> dict\n\n"
 "Return dictionary of the next log entry. Optional skip value will\n"
@@ -98,17 +220,19 @@ Journalctl_get_next(Journalctl *self, PyObject *args)
     const void *msg;
     size_t msg_len;
     const char *delim_ptr;
+    char bootid[33];
     PyObject *key, *value;
 
     SD_JOURNAL_FOREACH_DATA(self->j, msg, msg_len) {
         delim_ptr = memchr(msg, '=', msg_len);
 #if PY_MAJOR_VERSION >=3
         key = PyUnicode_FromStringAndSize(msg, delim_ptr - (const char*) msg);
-        value = PyUnicode_FromStringAndSize(delim_ptr + 1, (const char*) msg + msg_len - (delim_ptr + 1));
 #else
         key = PyString_FromStringAndSize(msg, delim_ptr - (const char*) msg);
-        value = PyString_FromStringAndSize(delim_ptr + 1, (const char*) msg + msg_len - (delim_ptr + 1));
 #endif
+        if (strncmp(msg, "_BOOT_ID", 8) == 0)
+            strncpy(bootid, (char *) delim_ptr + 1, 33);
+        value = Journalctl___process_field(self, key, delim_ptr + 1, (const char*) msg + msg_len - (delim_ptr + 1) );
         PyDict_SetItem(dict, key, value);
         Py_DECREF(key);
         Py_DECREF(value);
@@ -116,34 +240,35 @@ Journalctl_get_next(Journalctl *self, PyObject *args)
 
     uint64_t realtime;
     if (sd_journal_get_realtime_usec(self->j, &realtime) == 0) {
+        char realtime_str[20];
+        sprintf(realtime_str, "%llu", (long long unsigned) realtime);
+
 #if PY_MAJOR_VERSION >=3
-        value = PyUnicode_FromFormat("%llu", realtime);
+        key = PyUnicode_FromString("__REALTIME_TIMESTAMP");
 #else
-        value = PyString_FromFormat("%llu", realtime);
+        key = PyString_FromString("__REALTIME_TIMESTAMP");
 #endif
-        PyDict_SetItemString(dict, "__REALTIME_TIMESTAMP", value);
+        value = Journalctl___process_field(self, key, realtime_str, strlen(realtime_str));
+        PyDict_SetItem(dict, key, value);
+        Py_DECREF(key);
         Py_DECREF(value);
     }
 
-    char *bootid;
     sd_id128_t sd_id;
-#if PY_MAJOR_VERSION >=3
-    PyObject *temp;
-    temp = PyUnicode_AsASCIIString(PyDict_GetItemString(dict, "_BOOT_ID"));
-    bootid = PyBytes_AsString(temp);
-    Py_DECREF(temp);
-#else
-    bootid = PyString_AsString(PyDict_GetItemString(dict, "_BOOT_ID"));
-#endif
     if (sd_id128_from_string(bootid, &sd_id) == 0) {
         uint64_t monotonic;
         if (sd_journal_get_monotonic_usec(self->j, &monotonic, &sd_id) == 0) {
+            char monotonic_str[20];
+            sprintf(monotonic_str, "%llu", (long long unsigned) monotonic);
 #if PY_MAJOR_VERSION >=3
-            value = PyUnicode_FromFormat("%llu", monotonic);
+            key = PyUnicode_FromString("__MONOTONIC_TIMESTAMP");
 #else
-            value = PyString_FromFormat("%llu", monotonic);
+            key = PyString_FromString("__MONOTONIC_TIMESTAMP");
 #endif
-            PyDict_SetItemString(dict, "__MONOTONIC_TIMESTAMP", value);
+            value = Journalctl___process_field(self, key, monotonic_str, strlen(monotonic_str));
+
+            PyDict_SetItem(dict, key, value);
+            Py_DECREF(key);
             Py_DECREF(value);
         }
     }
@@ -151,11 +276,13 @@ Journalctl_get_next(Journalctl *self, PyObject *args)
     char *cursor;
     if (sd_journal_get_cursor(self->j, &cursor) > 0) { //Should return 0...
 #if PY_MAJOR_VERSION >=3
-        value = PyUnicode_FromString(cursor);
+        key = PyUnicode_FromString("__CURSOR");
 #else
-        value = PyString_FromString(cursor);
+        key = PyString_FromString("__CURSOR");
 #endif
-        PyDict_SetItemString(dict, "__CURSOR", value);
+        value = Journalctl___process_field(self, key, cursor, strlen(cursor));
+        PyDict_SetItem(dict, key, value);
+        Py_DECREF(key);
         Py_DECREF(value);
     }
 
@@ -195,11 +322,14 @@ Journalctl_add_match(Journalctl *self, PyObject *args)
     if (! PyArg_ParseTuple(args, "s#|s#", &match_key, &match_key_len, &match_value, &match_value_len))
         return NULL;
 
-    char *match;
+    void *match;
     int match_len;
     if (match_value) {
-        match = malloc(match_key_len + match_value_len + 2);
-        match_len = sprintf(match, "%s=%s", match_key, match_value);
+        match_len = match_key_len + 1 + match_value_len;
+        match = malloc(match_len);
+        memcpy(match, match_key, match_key_len);
+        memcpy(match + match_key_len, "=", 1);
+        memcpy(match + match_key_len + 1, match_value, match_value_len);
     }else{
         match = match_key;
         match_len = match_key_len;
@@ -234,9 +364,9 @@ Journalctl_add_matches(Journalctl *self, PyObject *args)
     PyObject *key, *value;
     while (PyDict_Next(dict, &pos, &key, &value)) {
 #if PY_MAJOR_VERSION >=3
-        if (!(PyUnicode_Check(key) && PyUnicode_Check(value))) {
+        if (!((PyUnicode_Check(key) || PyBytes_Check(key)) && (PyUnicode_Check(value) || PyBytes_Check(value)))) {
 #else
-        if (!(PyString_Check(key) && PyString_Check(value))) {
+        if (!((PyUnicode_Check(key) || PyString_Check(key)) && (PyUnicode_Check(value) || PyString_Check(value)))) {
 #endif
             PyErr_SetString(PyExc_ValueError, "Dictionary keys and values must be strings");
             return NULL;
@@ -337,19 +467,43 @@ Journalctl_seek(Journalctl *self, PyObject *args, PyObject *keywds)
 PyDoc_STRVAR(Journalctl_seek_realtime__doc__,
 "seek_realtime(realtime) -> None\n\n"
 "Seek to nearest matching journal entry to `realtime`. Argument\n"
-"`realtime` is an integer unix timestamp in usecs.");
+"`realtime` can be an integer unix timestamp in usecs or a "
+"datetime instance.");
 static PyObject *
 Journalctl_seek_realtime(Journalctl *self, PyObject *args)
 {
-    int64_t time;
-    if (! PyArg_ParseTuple(args, "L", &time))
+    PyObject *arg;
+    if (! PyArg_ParseTuple(args, "O", &arg))
         return NULL;
-    if (time < 0LL) {
-        PyErr_SetString(PyExc_ValueError, "Time must be positive integer");
+
+    uint64_t timestamp=-1LL;
+    if (PyDateTime_Check(arg)) {
+        PyObject *temp;
+        char *timestamp_str;
+        temp = PyObject_CallMethod(arg, "strftime", "s", "%s%f");
+#if PY_MAJOR_VERSION >=3
+        PyObject *temp2;
+        temp2 = PyUnicode_AsUTF8String(temp);
+        timestamp_str = PyBytes_AsString(temp2);
+        Py_DECREF(temp2);
+#else
+        timestamp_str = PyString_AsString(temp);
+#endif
+        Py_DECREF(temp);
+        timestamp = strtoull(timestamp_str, NULL, 10);
+    }else if (PyLong_Check(arg)) {
+        timestamp = PyLong_AsUnsignedLongLong(arg);
+#if PY_MAJOR_VERSION <3
+    }else if (PyInt_Check(arg)) {
+        timestamp = PyInt_AsUnsignedLongLongMask(arg);
+#endif
+    }
+    if ((int64_t) timestamp < 0LL) {
+        PyErr_SetString(PyExc_ValueError, "Time must be positive integer or datetime instance");
         return NULL;
     }
 
-    if (sd_journal_seek_realtime_usec(self->j, time) != 0) {
+    if (sd_journal_seek_realtime_usec(self->j, timestamp) != 0) {
         PyErr_SetString(PyExc_IOError, "Error seek to time");
         return NULL;
     }
@@ -357,29 +511,54 @@ Journalctl_seek_realtime(Journalctl *self, PyObject *args)
 }
 
 PyDoc_STRVAR(Journalctl_seek_monotonic__doc__,
-"seek_monotonic(monotonic) -> None\n\n"
+"seek_monotonic(monotonic[, bootid]) -> None\n\n"
 "Seek to nearest matching journal entry to `monotonic`. Argument\n"
-"`monotonic` is an integer timestamp from boot in usecs.");
+"`monotonic` is an integer timestamp from boot in usecs, or a\n"
+"timedelta instance.\n"
+"Argument `bootid` is a string representing which boot the\n"
+"monotonic time is reference to. Defaults to current bootid.");
 static PyObject *
 Journalctl_seek_monotonic(Journalctl *self, PyObject *args)
 {
-    int64_t time;
-    char *bootid;
-    if (! PyArg_ParseTuple(args, "Ls", &time, &bootid))
+    PyObject *arg;
+    char *bootid=NULL;
+    if (! PyArg_ParseTuple(args, "O|s", &arg, &bootid))
         return NULL;
-    if (time < 0LL) {
-        PyErr_SetString(PyExc_ValueError, "Time must be positive integer");
+
+    uint64_t timestamp=-1LL;
+    if PyDelta_Check(arg) {
+        PyObject *temp;
+        temp = PyObject_CallMethod(arg, "total_seconds", NULL);
+        timestamp = (uint64_t) (PyFloat_AsDouble(temp) * 1E6);
+        Py_DECREF(temp);
+    }else if (PyLong_Check(arg)) {
+        timestamp = PyLong_AsUnsignedLongLong(arg);
+#if PY_MAJOR_VERSION <3
+    }else if (PyInt_Check(arg)) {
+        timestamp = PyInt_AsUnsignedLongLongMask(arg);
+#endif
+    }
+
+    if ((int64_t) timestamp < 0LL) {
+        PyErr_SetString(PyExc_ValueError, "Time must be positive integer or timedelta instance");
         return NULL;
     }
 
     sd_id128_t sd_id;
-    if (sd_id128_from_string(bootid, &sd_id) == 0) {
-        if (sd_journal_seek_monotonic_usec(self->j, sd_id, time) != 0) {
-            PyErr_SetString(PyExc_IOError, "Error seek to time");
+    if (bootid) {
+        if (sd_id128_from_string(bootid, &sd_id) != 0) {
+            PyErr_SetString(PyExc_ValueError, "Invalid bootid");
             return NULL;
         }
     }else{
-        PyErr_SetString(PyExc_ValueError, "Invalid bootid");
+        if (sd_id128_get_boot(&sd_id) != 0) {
+            PyErr_SetString(PyExc_IOError, "Error getting current boot ID");
+            return NULL;
+        }
+    }
+
+    if (sd_journal_seek_monotonic_usec(self->j, sd_id, timestamp) != 0) {
+        PyErr_SetString(PyExc_IOError, "Error seek to time");
         return NULL;
     }
     Py_RETURN_NONE;
@@ -478,19 +657,22 @@ Journalctl_query_unique(Journalctl *self, PyObject *args)
     const void *uniq;
     size_t uniq_len;
     const char *delim_ptr;
-    PyObject *list, *value;
+    PyObject *list, *key, *value;
     list = PyList_New(0);
 
+#if PY_MAJOR_VERSION >=3
+    key = PyUnicode_FromString(query);
+#else
+    key = PyString_FromString(query);
+#endif
+
     SD_JOURNAL_FOREACH_UNIQUE(self->j, uniq, uniq_len) {
         delim_ptr = memchr(uniq, '=', uniq_len);
-#if PY_MAJOR_VERSION >=3
-        value = PyUnicode_FromStringAndSize(delim_ptr + 1, (const char*) uniq + uniq_len - (delim_ptr + 1));
-#else
-        value = PyString_FromStringAndSize(delim_ptr + 1, (const char*) uniq + uniq_len - (delim_ptr + 1));
-#endif
+        value = Journalctl___process_field(self, key, delim_ptr + 1, (const char*) uniq + uniq_len - (delim_ptr + 1));
         PyList_Append(list, value);
         Py_DECREF(value);
     }
+    Py_DECREF(key);
     return list;
 }
 #endif //def SD_JOURNAL_FOREACH_UNIQUE
@@ -576,6 +758,10 @@ Journalctl_this_machine(Journalctl *self, PyObject *args)
 }
 
 static PyMemberDef Journalctl_members[] = {
+    {"default_call", T_OBJECT_EX, offsetof(Journalctl, default_call), 0,
+    "default call for values for fields"},
+    {"call_dict", T_OBJECT_EX, offsetof(Journalctl, call_dict), 0,
+    "dictionary of calls for each field"},
     {NULL}  /* Sentinel */
 };
 
@@ -653,7 +839,7 @@ static PyTypeObject JournalctlType = {
     0,                                /* tp_dictoffset */
     (initproc)Journalctl_init,        /* tp_init */
     0,                                /* tp_alloc */
-    PyType_GenericNew,                /* tp_new */
+    Journalctl_new,                   /* tp_new */
 };
 
 #if PY_MAJOR_VERSION >= 3
@@ -675,6 +861,8 @@ initpyjournalctl(void)
 {
     PyObject* m;
 
+    PyDateTime_IMPORT;
+
     if (PyType_Ready(&JournalctlType) < 0)
 #if PY_MAJOR_VERSION >= 3
         return NULL;
